// author: Chuck Baker
// OPIE2 version 10.0

if (typeof(OPIE2) == 'undefined') var OPIE2 = {};

//		Common (global) variables
var Cc = Components.classes;
var Ci = Components.interfaces;
var Cu = Components.utils;

Cu.import("resource://gre/modules/AddonManager.jsm");
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://opie2/taffy.js");

//let consoleJSM = Cu.import("resource://gre/modules/Console.jsm", {});
//let OPIE2console = consoleJSM.console;
//let OPIE2console = Cu.import("resource://gre/modules/Console.jsm", {}).console;


window.addEventListener("load",function(){OPIE2.init()},true);

OPIE2 = {
	InstalledExtsDb: TAFFY(),			// db of all installed extensions
	ExtPrefsDb: TAFFY(),					// db of OPIE2 extension preferences
	FxPrefsDb: TAFFY(),						// db of all Fx  preferences (as in about:config)
	InstalledExts: [],						// Array of installed extension objects
	UserDefinedNamesDb: TAFFY(),	// db of user-defined package names
	ExtensionStatusDb: TAFFY(),		// db of extensions OPIE2 has prefs for
	opie2ProfDir: {}, 						// Pointer to profile directory
	opie2ExtList: [],
	opie2FxPrefList: [],
	opie2UDList: [],
	topWin: Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator).getMostRecentWindow("navigator:browser"),  
	opie2InstalledExts: [],
	opie2ExtNames: [],						// List of extensions
	opie2ProfName: "",						// Current profile name
	opie2RootEdits: {},						// List of edited preference roots
	OPIE2_GUID: "OPIE2@guid.customsoftwareconsult.com",
	resetObj: {},									// Used for 'Reset prefs' button
	opie2Msg: [],									// Opie messages
	opie2ImportList: [],					// List of prefs to import
	opie2ExportDir: "",						// Export directory path
	opie2SelectionCount: 0,				// Number of items selected for export
	opie2ExtensionCount: 0,				// Number of extensions avaliable for export
	opie2FileUtils: {},						// Pointer to "resource://gre/modules/FileUtils.jsm"
	opie2TmpDir: {},							// Temp dir used to extract pref files
	opie2Overwrite: false,				// Overwrite existing exports?
	opie2ErrCnt: 0,								// Count of locking errors
	opie2Err: 0,									// Error flag (0 - 4)
	opie2ExtDir: {},							// Pointer to the Fx profile extension directory
	opie2Win: {},									// Pointer to the currently open xul window

	opie2Prefs: Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch(""),

	opie2GetUnicharPref: function(prefName) {
		try{
			return OPIE2.opie2Prefs.getComplexValue(prefName,  Ci.nsISupportsString).data;
		}catch(err){
			alert(prefName+"\n\n"+err);
		}
		return true;
	},

	opie2SetUnicharPref: function(prefName, prefValue) {
		let supportsString = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
		supportsString.data = prefValue;
		this.opie2Prefs.setComplexValue(prefName, Ci.nsISupportsString, supportsString);
		return true;
	},

	opie2PrefObj: function(Type,Name,Value){
		this.Type = Type;
		this.Name = Name;
		this.Value = Value;
	},

	opie2AlertObj: function(Text,Style){
		this.Text = Text;
		this.Style = Style;
		return true;
	},

	init: function (){
		// Initialize variables and objects
		window.removeEventListener("load",OPIE2.init,true);
		let windowtype = document.documentElement.getAttribute("windowtype");
		if(windowtype != "navigator:browser") return true;
		this.GetInstalledExtensionsList();
		this.opie2SetMsgs();
		this.opie2ExportDir = this.opie2GetUnicharPref("extensions.opie2.exportDir");
		this.opie2FileUtils = Cu.import("resource://gre/modules/FileUtils.jsm", {}).FileUtils;
		
		this.opie2ExtDir = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
		let tmp = this.opie2ExtDir.leafName;
		let p = tmp.indexOf(".")
		this.opie2ProfName  = tmp.substr(p+1);
		OPIE2.storage.setItem('opie2ProfName',this.opie2ProfName);
		this.opie2ExtDir.append('extensions');
		
		return true;
	},
	
	storage: {
		// This object replaces the deprecated "Application.storage" method
		setItem: function setItem(key, value){
			if(typeof OPIE2.topWin.storage == "undefined") OPIE2.topWin.storage = {};
			if(key == null || key == "") return false;
			OPIE2.topWin.storage[key] = value;
			return true;
		},
		getItem: function getItem(key){
			if(key == null || key == "") return false;
			if(typeof OPIE2.topWin.storage[key] == "undefined") return false;
			let value = OPIE2.topWin.storage[key];
			return value;
		},
	},
	
	init2: function (){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status
		
		// Build InstalledExtDb database
		let data = this.storage.getItem('opie2InstalledExts');
		let objArray = JSON.parse(data);
		this.InstalledExtsDb = TAFFY();
		for(let i in objArray){
			let obj = JSON.parse(objArray[i]);
			this.InstalledExtsDb.insert(obj);
		}
		this.readData();
		
		// Add user-defined names to the list
		let udn = this.UserDefinedNamesDb().get();
		for(let i in udn){
			let name = udn[i].udname;
			let item = {};
			let datafileName = this.opie2ReplaceSpaces(name);
			datafileName += "("+this.opie2Msg[43]+")";
			this.opie2ProfName = OPIE2.storage.getItem('opie2ProfName');
			datafileName += "{" + this.opie2ReplaceSpaces(this.opie2ProfName) + "}" + ".oprefs";
			item.Guid = 'userdefined';
			item.Name = name;
			item.DatafileName = datafileName;
			item.Version = "";
			item.isXpi = false;
			this.InstalledExtsDb.insert(item);
		}
		
		this.GetAllFxPreferences();	
					
		this.opie2ExportDir = this.opie2GetUnicharPref("extensions.opie2.exportDir");
		this.opie2FileUtils = Cu.import("resource://gre/modules/FileUtils.jsm", {}).FileUtils;
		
		this.opie2ExtDir = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
		this.opie2ExtDir.append('extensions');	
		
		// Pre-set the reset choices
		let type1 = this.opie2Prefs.getBoolPref("extensions.opie2.type1");	
		let type2 = this.opie2Prefs.getBoolPref("extensions.opie2.type2");		
		let type3 = this.opie2Prefs.getBoolPref("extensions.opie2.type3");
		let type4 = this.opie2Prefs.getBoolPref("extensions.opie2.type4");		
		
		document.getElementById('type1').setAttribute('checked',type1);
		document.getElementById('type2').setAttribute('checked',type2);
		document.getElementById('type3').setAttribute('checked',type3);
		document.getElementById('type4').setAttribute('checked',type4);
		return true;
	},
	
	GetAllFxPreferences: function(){
		let prefService = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService); 
		let prefBranch = prefService.getBranch(null); 
		let obj = {};
		let prefs = prefBranch.getChildList("",obj)
		for(let i in prefs){
			let pref = prefs[i];
			let type = prefBranch.getPrefType(pref);
			let value = this.opie2GetExtPref(pref,type);
			let item = {};
			let Type = "";
			switch (type){
				case prefBranch.PREF_BOOL:	
					Type = "boolean";
					break;
				case prefBranch.PREF_INT:
					Type = "integer";
					break;
				case prefBranch.PREF_STRING:
					Type = "string";
					break;
			}
			
			item.name = pref;
			item.type = Type;
			item.value = value;
			OPIE2.FxPrefsDb.insert(item);	
		}
	},
	
	GetInstalledExtensionsList: function (){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		// Get a list of all installed extensions that have default preferences (as determined by the existance of optionsURL
		AddonManager.getAddonsByTypes(["extension"],function(list) {
			let cnt = 0;
			let dataArray = [];
			list.forEach(function(addons) {
				AddonManager.getAddonByID(addons.id, function(addon) {
					cnt++;
					if(addon.optionsURL){	// If true, assume preferences exist
						let datafileName = OPIE2.opie2StripSpaces(addons.name)+"{" + addons.version + "}" + ".oprefs";
						let item = {};
						item.Guid = addons.id;
						item.Name = addons.name;
						item.DatafileName = datafileName;
						item.Version = addons.version;
						item.isXpi = OPIE2.setIsXpi(item);
						
						let p = addon.optionsURL.indexOf("chrome://");
						let chromeName = "";
						if(p < 0){
							chromeName = OPIE2.opie2Msg[13];	// "Unknown"
						}else{
							chromeName = addon.optionsURL.substr(p+9);
							chromeName = chromeName.match(/^\w+/).toString();
						}
						item.chromename = chromeName;
						dataArray.push(JSON.stringify(item))
					}
					if(cnt == list.length){
						// Save results so it can be loaded from any window
						OPIE2.storage.setItem('opie2InstalledExts',JSON.stringify(dataArray));
					}
				});
			});
		});
		return true;
	},

	opie2SetMsgs: function (){
		// Error messages, prompts, etc.
		if(!this.opie2Msg[0]){
			let bundle = Services.strings.createBundle("chrome://opie2/locale/opie2.properties");
			let enumerator = bundle.getSimpleEnumeration();
			while (enumerator.hasMoreElements()) {
				let string = enumerator.getNext().QueryInterface(Ci.nsIPropertyElement);
				let stringName = string.key;
				if(stringName.substr(0,3) != "msg") continue;
				let index = stringName.substr(3);
				index = parseInt(index);
				let value = string.value;
				this.opie2Msg[index] = value;
			}
		}	
		return true;
	},

	opie2StripSpaces: function(extName){
		// Remove spaces and undesirable characters from extension name
		let tmp = ""; 
		let mask = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"; 
		mask += ".+-_";
		for (let i = 0; i <= extName.length;  i++){
			let c = extName.charAt(i); 
			if( mask.indexOf(c) != -1 ){tmp += c};
		}
		return tmp;
	},
	
	opie2ReplaceSpaces: function(extName){
		// Remove spaces and undesirable characters from extension name
		let name = extName;
		let tmp = name.replace(/ /g,"_");
		let retValue = this.opie2StripSpaces(tmp);
		return retValue;
	},

	getOpieVersion: function () {
		AddonManager.getAddonByID(this.OPIE2_GUID, function(addon) {
			OPIE2.opie2SetMsgs();
			document.getElementById("opie2VersionID").value = OPIE2.opie2Msg[0] + ": " + addon.version;
		});
		return true;
	},
	
	writeData: function(){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		let dFile = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
		let dataFileName = "OPIE2data.json";
		dFile.append(dataFileName);
		if(!dFile.exists()) dFile.create(Ci.nsIFile.NORMAL_FILE_TYPE, 0x1ED);	// 0x1ED = 0755, 0x1FF = 0777
		let fos = Cc["@mozilla.org/network/file-output-stream;1"].createInstance(Ci.nsIFileOutputStream);
		fos.init(dFile, 0x02 | 0x08 | 0x20, 0x1ED, 0); // write, create, truncate
		
		let cos = Cc["@mozilla.org/intl/converter-output-stream;1"].createInstance(Ci.nsIConverterOutputStream);
		cos.init(fos, "UTF-8", 0, 0x0000);
		let prefs = this.ExtPrefsDb().order('ename,pname').get();
		let header1 = "<<OPIE2DATA>>";
		let header2 = "<<OPIE2USERDEFINED>>";
		let header3 = "<<OPIE2EXTSTATUS>>";
		
		// Write contents of ExtPrefsDb to file in "<<OPIE2DATA>>" section
		cos.writeString(header1+"\n");
		for(let i in prefs){
			let pref = prefs[i];
			let item = {};
			item.ename = pref.ename;
			item.guid = pref.guid;
			item.pname = pref.pname;
			item.ptype = pref.ptype;
			cos.writeString(JSON.stringify(item)+"\n"); 
		};
		
		// Write contents of UserDefinedNamesDb to file in "<<OPIE2USERDEFINED>>" section
		cos.writeString(header2+"\n");
		let udnames = this.UserDefinedNamesDb().order('udname').get();
		for(let i in udnames){
			let name = udnames[i].udname;
			let item = {};
			item.udname = name;
			cos.writeString(JSON.stringify(item)+"\n"); 
		};
		
		// Write contents of ExtensionStatusDb to file in "<<<<OPIE2EXTSTATUS>>>>" section
		cos.writeString(header3+"\n");
		let exts = this.ExtensionStatusDb().order('name').get();
		let ud = "undefined";
		for(let i in exts){
			let item = {};
			item.name = exts[i].name;
			item.guid = exts[i].guid;
			item.status = exts[i].status;
			if(typeof item.name == ud || typeof item.guid == ud) continue; // ExtensionStatusDb: guid, name, status
			cos.writeString(JSON.stringify(item)+"\n"); 
		};
		cos.close();
		fos.close();
		return true;
	},
	
	readData: function(){
		// Read contents of file into ExtPrefsDb
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		let dFile = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
		let dataFileName = "OPIE2data.json";
		dFile.append(dataFileName);
		if(!dFile.exists() || dFile.fileSize == 0){
			this.noData();	
			return true;
		}
		let fis = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
		fis.init(dFile, 0x01, 0x124, 0);
		fis.QueryInterface(Ci.nsILineInputStream);

		let cis = Cc["@mozilla.org/intl/converter-input-stream;1"].createInstance(Ci.nsIConverterInputStream);
		cis.init(fis,"UTF-8", 0, 0x0000);
		let header;
		let header1 = "<<OPIE2DATA>>";
		let header2 = "<<OPIE2USERDEFINED>>";
		let header3 = "<<OPIE2EXTSTATUS>>";
		
		let lis = cis.QueryInterface(Ci.nsIUnicharLineInputStream);
		let line = {}, hasmore;	
		do {
			hasmore = lis.readLine(line);	
			let item = line.value;
			if(item == header1){
				hasmore = lis.readLine(line);	
				item = line.value;
				header = header1;
			}
			if(item == header2){
				hasmore = lis.readLine(line);	
				item = line.value;
				header = header2;
			}
			if(item == header3){
				hasmore = lis.readLine(line);	
				item = line.value;
				header = header3;
			}
			let ud = "undefined";
			if(typeof item == ud || item == "") continue;
			switch(header){
				case header1:	
					let pref = JSON.parse(item);
					if(typeof pref.ename == ud || typeof pref.guid == ud || typeof pref.pname == ud || typeof pref.ptype == ud) break;
					this.ExtPrefsDb.insert(pref);			// ExtPrefsDb: ename, guid, pname, ptype
					break;
				case header2:
					let udname = JSON.parse(item);
					if(typeof udname.udname == ud) break;
					this.UserDefinedNamesDb.insert(udname);	// UserDefinedNamesDb: udname			
					break;;
				case header3:
					let rec = JSON.parse(item);
					if(typeof rec.name == ud || typeof rec.guid == ud || typeof rec.status == ud) break;
					this.ExtensionStatusDb.insert(rec);	// ExtensionStatusDb: guid, name, status			
					break;;
			}
		} while(hasmore);
		cis.close();
		fis.close();
		let cnt = this.ExtPrefsDb().count();
		if(cnt == 0) this.noData();
		return true;
	},
	
	noData: function(){
		let tmp = this.opie2Msg[46]+"\n";
		tmp += this.opie2Msg[47]+"\n";
		tmp += this.opie2Msg[48];
		this.opie2Alert(tmp);
		return false;
	},

	opie2SelFiles: function (){
		// Select files to import
		const nsIFilePicker = Ci.nsIFilePicker;
		let fp = Cc["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);
		let win = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator).getMostRecentWindow("opie2:main");
		fp.init(win, this.opie2Msg[7], nsIFilePicker.modeOpenMultiple);
		let filter = "*.oprefs"; 
		fp.appendFilter(filter,filter);
		
		// Set default directory to last import directory if it is set, otherwise try the export directory
		let err = Boolean(false); 
		let dd = Cc["@mozilla.org/file/local;1"]
						 .createInstance(Ci.nsIFile);
		try{
			dd.initWithPath(this.opie2GetUnicharPref("extensions.opie2.importDir"));
		}catch(e){
			try{
				dd.initWithPath(this.opie2GetUnicharPref("extensions.opie2.exportDir"));
			}catch(e){err = true;}
		}
		if(!err){fp.displayDirectory = dd;}
		fp.open(function(rv){
			OPIE2.opie2ImportList = [];
			let fnames = []; 
			if (rv == nsIFilePicker.returnOK){
				let files = fp.files; 
				while(files.hasMoreElements()) {
					var file = files.getNext()
						.QueryInterface(Ci.nsIFile);
					OPIE2.opie2ImportList.push(file.path);
					fnames.push(file.leafName);
				}
			}
			document.getElementById("opie2ImportFileID").value = fnames.join("\n");
			let importButton = document.getElementById("opie2ImportButton"); 
			if(fnames.length == 0){
				importButton.disabled = true;
			}else{
				importButton.disabled = false;
			}	
		});
		return true;
	},

	opie2SelDir: function (){
		// Select export directory
		const nsIFilePicker = Ci.nsIFilePicker;
		let fp = Cc["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);
		let win = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator).getMostRecentWindow("opie2:main");
		fp.init(win, this.opie2Msg[4], nsIFilePicker.modeGetFolder);
		fp.appendFilters(nsIFilePicker.filterAll | nsIFilePicker.filterText);
		
		let extDir = ""; 

		// Set default directory to current export directory
		let prefName = "extensions.opie2.exportDir"; 
		if(this.opie2Prefs.prefHasUserValue(prefName)){
			extDir = this.opie2GetUnicharPref(prefName);
			if(extDir != ""){
				let dd = Cc["@mozilla.org/file/local;1"]
								 .createInstance(Ci.nsIFile);
				let err = Boolean(false); 
				try {
					dd.initWithPath(extDir);
				}catch(e){err = true;}
				if(!err){fp.displayDirectory = dd;}
			}
		}
		fp.open(function(rv){
			if(rv == nsIFilePicker.returnOK){
				rv = fp.file;
				extDir = rv.path;
				OPIE2.opie2SetUnicharPref("extensions.opie2.exportDir", extDir);
				document.getElementById("opie2ExportDirID").setAttribute("value",extDir);
				OPIE2.buttonCheck();
			}
		});
		return true;
	},

	opie2PlaySound: function(which){
		if(!this.opie2Prefs.getBoolPref("extensions.opie2.playSounds")) return true;
		let soundURL = "chrome://opie2/content/sounds/"+which+".ogg"; 
		let audio = new Audio(soundURL); 
		audio.play();
		return true;
	},

	logMsg: function(msg) {
		msg = "OPIE2: "+msg;
		OPIE2console.log(msg)
		let cs = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);
		cs.logStringMessage(msg);
		return true;
	},
		
	propsConsole: function(obj,title,ignoreProperties,ignoreFunctions,ignoreObjects,expandObjs,sort){
		if(title != null){
			title = "("+title+")";
		}else{
			title = "";
		}
		if (obj === null) {
			this.logMsg("propsConsole called with null argument: "+title+"\n");
			return false;
		}
		if (obj === undefined) {
			this.logMsg("propsConsole called with undefined argument: "+title+"\n");
			return false;
		}
		let tmp, line, pVal, type, msg, attrb = [], func = [], objects = [];
		msg = "propsConsole: "+title+"\ntypeof '"+obj+"': ("+typeof obj+"), "+((sort) ? "Sorted" : "Unsorted")+"\n";
		for(let property in obj){
			line = property;
			try{
				pVal = obj[property];
			}catch(err){
				pVal = "(Unreachable)  "+err;
			}
			try{
				type = typeof obj[property];
			}catch(e){
				type = "Unknown";
			}
			try{
				tmp = line + " = "+pVal;
			}catch(e){
				tmp = "(Unreachable)";
			}
			switch(type){
				case 'function':
					if(!ignoreFunctions) func.push(tmp);
					break;
				case 'object':
					if(ignoreObjects) continue;
					try{
						let props = JSON.stringify(obj[property]);
						if(expandObjs && props != "null") tmp += " --> "+props;
					}catch(err){;;}
					objects.push(tmp);
					break;
				default:
					if(!ignoreProperties) attrb.push(tmp);
					break;
			}
		} 
		if(func.length == 0) func.push("(none)");
		if(objects.length == 0) objects.push("(none)");
		if(attrb.length == 0) attrb.push("(none)");

		if(sort){
			attrb.sort();
			func.sort();
			objects.sort();
		}

		if(!ignoreProperties){
			msg += "\n*** PROPERTIES/ATTRIBUTES ***\n";
			for(let i in attrb){
				msg += attrb[i]+"\n";
			}
		}
		
		if(!ignoreFunctions){
			msg += "\n*** FUNCTIONS ***\n";
			for(let i in func){
				msg += func[i]+"\n";
			}
		}
		
		if(!ignoreObjects){
			msg += "\n*** OBJECTS ***\n";
			for(let i in objects){
				msg += objects[i]+"\n";
			}
		}
		this.logMsg(msg);
		return true;
	},
	
	propsDb: function(dbName){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status
		
		OPIE2.logMsg("propsDb: "+dbName+", count: "+OPIE2[dbName]().count());
		let pArray = OPIE2[dbName]().get();
		for(let i in pArray){
			let rec = pArray[i];
			OPIE2.propsConsole(rec,"rec",false,true,true,true,true)
		}
	},

	opie2Open: function (){
		this.opie2Win = window.openDialog('chrome://opie2/content/opie2.xul', 'OPIE2', 'chrome,titlebar,toolbar,centerscreen,resizable');	
		return true;
	},

	exportPane: function (){
		document.getElementById("opie2button2ID").disabled = false;
		document.getElementById("opie2button3ID").disabled = false;
		document.getElementById("opie2button4ID").disabled = false;
		this.clearExportList();
		this.buildExportList();
		return true;
	},

	opie2Import: function (){
		let opie2Results = []; 
		let n = 0; 
		if(this.opie2ImportList == ""){
			let style =  "<style>color: red;font-size: 10pt;</style>";
			let tmp = style+document.getElementById("opie2.import.file.default").value; 
			this.opie2Alert(tmp);	// No extensions selected
			return false;
		}
		for(var i in this.opie2ImportList){
			let prefFile = this.opie2ImportList[i]; 
			n++;

			// Read the data
			let dataFile = Cc["@mozilla.org/file/local;1"]
											 .createInstance(Ci.nsIFile);
			dataFile.initWithPath(prefFile);
			
			// open an input stream from file
			let istream = Cc["@mozilla.org/network/file-input-stream;1"]
										.createInstance(Ci.nsIFileInputStream);
			istream.init(dataFile, 0x01, 0444, 0);
			istream.QueryInterface(Ci.nsILineInputStream);
		
			let cis = Cc["@mozilla.org/intl/converter-input-stream;1"]
								.createInstance(Ci.nsIConverterInputStream);
		
			let charset = "UTF-8"; 
			cis.init(istream,charset, 1024, Ci.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
			let lis = cis.QueryInterface(Ci.nsIUnicharLineInputStream); 

			// read lines into array and parse into individual fields
			let line = {}, hasmore, pcnt = 0; 
			do {
				hasmore = lis.readLine(line);
				let tmp = line.value; 
				if(tmp.length == 0){break;}
				
				// Write the preferences
				let pref = JSON.parse(tmp);   
				switch(pref.Type){
					case "boolean":
						this.opie2Prefs.setBoolPref(pref.Name,Boolean(pref.Value)); 
						break;
					case "integer":
						this.opie2Prefs.setIntPref(pref.Name,parseInt(pref.Value));
						break;
					case "string":
						this.opie2SetUnicharPref(pref.Name, String(pref.Value));
						break;
				}
				pcnt++;
			} while(hasmore);
			
			opie2Results.push(dataFile.leafName+":\t"+pcnt+" "+this.opie2Msg[8])

			istream.close();
			cis.close();	
		}
		
		opie2Results.push(" ");
		opie2Results.push(this.opie2Msg[9].replace(/%NUM%/,n));
		document.getElementById("opie2ImportsResultsID").value = opie2Results.join("\n");
		this.opie2PlaySound('success');
		return true;
	},

	opie2Select: function(bool){
		// Select/deselect all
		for(let i in this.opie2ExtNames){
			let id = "expList:"+this.opie2ExtNames[i];
			let item = document.getElementById(id); 
			item.setAttribute("checked",bool);
		}
		if(bool){
			this.opie2SelectionCount = this.opie2ExtensionCount;
		}else{
			this.opie2SelectionCount = 0;
		}
		this.setExportCaption();
	},
		
	InitAllPrefs: function(){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		let pmbox = document.getElementById('opie2progressmeter');
		pmbox.hidden = false;
		let pm = document.getElementById('opie2initPM');
		document.getElementById('opie2ExtListCntBox').setAttribute('hidden', true);
		document.getElementById('opie2InitAllTree').hidden = true;	
		document.getElementById('opie2ExtListCntBox').hidden = true;			
		document.getElementById('expandall').hidden = true;
		document.getElementById('collapseall').hidden = true;
		let includeUD = document.getElementById("includeUDchkbox").checked;

		this.pauseThread();
		
		// Clear the db
		if(includeUD){
			this.ExtPrefsDb().remove();
		}else{
			this.ExtPrefsDb({guid:{'!is':'userdefined'}}).remove();
		}
		this.ExtensionStatusDb = TAFFY();
		
		let typeAll = [];
		let exts = this.InstalledExtsDb().order("Name").get();	// An array of objects sorted by name
		let tot = exts.length;
		let cnt = 0;	
		for(let i in exts){
			cnt++;
			let pct = (cnt/tot)*100;
			pm.value = pct;
			document.getElementById('progresscount').value = parseInt(pct,10)+"%";
			let rec = exts[i];
			if(rec.Guid != 'userdefined'){
				let type1 = this.GetType1(rec);
				this.pauseThread();
				let type2 = this.GetType2(rec);
				this.pauseThread();
				let type3 = this.GetType3(rec);
				this.pauseThread();
				let type4 = this.GetType4(rec);
				this.pauseThread();
				typeAll = typeAll.concat(type1,type2,type3,type4);
				
				for(let i in typeAll){
					let item = typeAll[i];
					if(this.ExtPrefsDb(item).count() == 0) this.ExtPrefsDb.insert(item);
					
					let cnt = this.ExtensionStatusDb({name:item.ename}).count();
					if(cnt == 0){
						let obj = {};
						obj.guid = item.guid;
						obj.name = item.ename;
						obj.status = 1;
						this.ExtensionStatusDb.insert(obj);
					}
				}
			}else{
				if(!includeUD){
					// See if user-defined package has prefs assigned
					let cnt2 = this.ExtensionStatusDb({name:rec.Name}).count();
					if(cnt2 != 0){
						let obj = {};
						obj.guid = rec.Guid;
						obj.name = rec.Name;
						obj.status = 2;
						this.ExtensionStatusDb.insert(obj);
					}
				}
			}
		}	
		
		setTimeout(function(){
			document.getElementById('opie2progressmeter').hidden = true;
			OPIE2.opie2BuildInitAllTree();
		},500);

		return true;
	},
		
	opie2BuildInitAllTree: function (){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status
		
		let theList = document.getElementById("opie2InitAllChildren");
		
		// Remove the existing tree and rebuild it
		let children = theList.childNodes;
		let n = children.length;
		for (let i = 0; i < n; i++) {
			theList.removeChild(children[0]);
		}
	
		let db = this.ExtPrefsDb().order("ename").get();
		let item,row,col,cell,erow,rec,eInfo,eItem,echildren,pitem,prow,pcell,prefs,pref
		let cGuid = "";
		let ecnt = pcnt = ptcnt = pscnt = picnt = pbcnt = 0;
		for(let i in db){
			item = db[i];
			if(item.guid == cGuid) continue;
			ecnt++;
			
			// Write previous header if this is not the first loop iteration
			if(cGuid != "") theList.appendChild(eitem);
			cGuid = item.guid;
			
			// New header
			eitem = document.createElement('treeitem');
			eitem.setAttribute('container', "true");
			rec = this.InstalledExtsDb({Guid:cGuid}).get();
			eInfo = rec[0];
						
			erow = document.createElement('treerow');
			ecell = document.createElement('treecell');
			ecell.setAttribute('label',eInfo.Name);
			ecell.setAttribute('value',eInfo.Name);
			erow.appendChild(ecell);
			
			ecell = document.createElement('treecell');
			ecell.setAttribute('label',eInfo.Version);
			ecell.setAttribute('value',eInfo.Version);
			erow.appendChild(ecell);
			
			ecell = document.createElement('treecell');
			let guid = eInfo.Guid;
			if(guid == 'userdefined') guid = this.opie2Msg[43];
			ecell.setAttribute('label',guid);
			ecell.setAttribute('value',guid);
			erow.appendChild(ecell);
			eitem.appendChild(erow);
		
			echildren = document.createElement('treechildren');
			prefs = this.ExtPrefsDb({guid:cGuid}).order("pname").get();
			for(let j in prefs){
				pref = prefs[j];
				pitem = document.createElement('treeitem');
				pitem.setAttribute('container', "false");
				prow = document.createElement('treerow');
				pcell = document.createElement('treecell');
				pcell.setAttribute('label',pref.pname);
				pcell.setAttribute('value',pref.pname);
				prow.appendChild(pcell);
				
				pcell = document.createElement('treecell');
				pcell.setAttribute('label',pref.ptype);
				pcell.setAttribute('value',pref.ptype);
				prow.appendChild(pcell);
				
				pcell = document.createElement('treecell');
				let ext = this.FxPrefsDb({name:pref.pname}).get();
				let value = ext[0].value;
				pcell.setAttribute('label',value);
				pcell.setAttribute('value',value);
				prow.appendChild(pcell);
								
				pitem.appendChild(prow);
				echildren.appendChild(pitem);	
				eitem.appendChild(echildren);
				
				pcnt++;		// Count of prefs for this extension
				ptcnt++;	// Grand total count of all preferences
				if(pref.ptype == "string") pscnt++;
				if(pref.ptype == "integer") picnt++;
				if(pref.ptype == "boolean") pbcnt++;
			}
			
			// Add preference totals
			pitem = document.createElement('treeitem');
			pitem.setAttribute('container', "false");
			prow = document.createElement('treerow');
			pcell = document.createElement('treecell');
			
			let txt = this.opie2Msg[42];
			txt = txt.replace("%TOT%", ptcnt);
			txt = txt.replace("%STR%", pscnt);
			txt = txt.replace("%INT%", picnt);
			txt = txt.replace("%BOOL%", pbcnt);
			pcell.setAttribute('label',txt);
			prow.appendChild(pcell);
						
			pitem.appendChild(prow);
			echildren.appendChild(pitem);	
			eitem.appendChild(echildren);
			ptcnt = pscnt = picnt = pbcnt = 0;
			
			theList.appendChild(eitem);
		}
		document.getElementById('opie2ExtListCnt').setAttribute('value', ecnt);
		document.getElementById('opie2ExtPrefCnt').setAttribute('value', pcnt);
		document.getElementById('opie2ExtListCntBox').setAttribute('hidden', false);
		document.getElementById('opie2InitAllTree').hidden = false;	
		document.getElementById('opie2ExtListCntBox').hidden = false;			
		document.getElementById('expandall').hidden = false;
		document.getElementById('collapseall').hidden = false;
		return true;
},
	
	GetType1: function(rec){
		// Get the prefs of type 'extensions.{chromename}'
		retArray = [];
		let key = "extensions."+rec.chromename+".";
		let prefs = OPIE2.FxPrefsDb().filter({name:{left:key}}).each(function(entry,num){
			let pref = {};
			pref.ename = rec.Name;
			pref.guid = rec.Guid;
			pref.pname = entry.name;
			pref.ptype = entry.type;
			retArray.push(pref);
		});
		return retArray;
	},
	
	GetType2: function(rec){
		// Get the prefs of type 'extensions.{guid}'
		retArray = [];
		let key = "extensions."+rec.Guid+".";
		let prefs = OPIE2.FxPrefsDb().filter({name:{left:key}}).each(function(entry,num){
			let pref = {};
			pref.ename = rec.Name;
			pref.guid = rec.Guid;
			pref.pname = entry.name;
			pref.ptype = entry.type;
			retArray.push(pref);
		});
		return retArray;
	},
	
	GetType3: function(rec){
		// Get the prefs of type 'chromename.'
		retArray = [];
		let key = rec.chromename+".";
		let prefs = OPIE2.FxPrefsDb().filter({name:{left:key}}).each(function(entry,num){
			let pref = {};
			pref.ename = rec.Name;
			pref.guid = rec.Guid;
			pref.pname = entry.name;
			pref.ptype = entry.type;
			retArray.push(pref);
		});
		return retArray;
	},

	GetType4: function(rec){		
		// Get the preference file from 'defaults/preferences/'.  Extract the file to a tmp dir if the extension is an xpi
		let entry, file, entries, zipR, prefFile, outFile, retArray = [], curExtDir, xpiFile, id = rec.Guid;
		if(rec.isXpi){
			// Extract all the prefs files into a tmp directory
			if(!this.readyTmpDir()) return false;
			curExtDir = this.opie2ExtDir.clone();
			curExtDir.append(id+".xpi");
			xpiFile = curExtDir;
			
			// Unzip the xpi's preference files
			zipR = Cc["@mozilla.org/libjar/zip-reader;1"].createInstance(Ci.nsIZipReader);
			zipR.open(xpiFile);

			entries = zipR.findEntries("defaults/preferences/*.js$");
			while (entries.hasMore()) {
				entry = entries.getNext();
				prefFile = entry.substr(entry.lastIndexOf("/")+1);
				outFile = this.opie2TmpDir.clone();
				outFile.append(prefFile);
				zipR.extract(entry,outFile);
				outFile.permissions = 0x1ff;
				this.pauseThread();
			}
			zipR.close();		
			curExtDir	= this.opie2TmpDir;
		}else{
			curExtDir = this.opie2ExtDir.clone();
			curExtDir.append(id);
			curExtDir.append('defaults');
			curExtDir.append('preferences');
		}
		if(!curExtDir.exists()) return false;
		
		// Parse the dir for all pref files and extract the prefs
		entries = curExtDir.directoryEntries;
		while(entries.hasMoreElements()){
			this.pauseThread();
			entry = entries.getNext();
			entry.QueryInterface(Ci.nsIFile);
			file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
			file.initWithPath(entry.path);
			let fis = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
			fis.init(file, 0x01, 0x124, 0);
			fis.QueryInterface(Ci.nsILineInputStream);

			let cis = Cc["@mozilla.org/intl/converter-input-stream;1"].createInstance(Ci.nsIConverterInputStream);
			cis.init(fis,"UTF-8", 0, 0x0000);
			
			let lis = cis.QueryInterface(Ci.nsIUnicharLineInputStream);
			let line = {}, hasmore;	
			do {
				hasmore = lis.readLine(line);			
				let prefLine = line.value.replace(/'/g,'"');	// Replace single quotes to double quotes
				if(prefLine.substr(0,6) != 'pref("') continue;
				let match, pref;
				pref = prefLine.substr(6);
				let quotePos = pref.indexOf('"');
				let prefname = pref.substr(0,quotePos);
				
				// FxPrefsDb:  name, type, value
				let item = this.FxPrefsDb({name:prefname}).get();
				if(typeof item[0] != 'undefined'){
					pref = {};
					pref.ename = rec.Name;
					pref.guid = rec.Guid;
					pref.pname = item[0].name;
					pref.ptype = item[0].type;
					retArray.push(pref);
				}
			} while(hasmore);
			cis.close();
			fis.close();
		}
		return retArray;
	},
	
	initConfigMenu:	function(){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		// Build extension list
		let theList = document.getElementById("configExtPopupID"); 
		// Remove the existing list and rebuild it
		let children = theList.childNodes;
		let n = children.length;
		for (let i = n-1; i >= 0; i--){
			let node = children[i];
			if(node.tagName == 'menuitem') theList.removeChild(node);
		}

		let exts = this.InstalledExtsDb().order("Name").get();	// An array of objects sorted by name
		for(let i in exts){
			let ext = exts[i];
			if(ext.Guid == "userdefined") continue;
			let item = document.createElement('menuitem');
			item.setAttribute('label',ext.Name);
			item.setAttribute('value',ext.Name);
			item.addEventListener('click', function(event){OPIE2.dispExtInfo(event,this)}, false);
			let cnt = this.ExtPrefsDb({ename:ext.Name}).count();
			let klass, status;
			if(cnt != 0){
				let cnt2 = this.ExtensionStatusDb({name:ext.Name}).count();
				if(cnt2 == 0){
					let obj = {};
					obj.name = name;
					obj.guid = "userdefined";
					obj.status = 2;
					this.ExtensionStatusDb.insert(obj);	
				}
				let statusRec = this.ExtensionStatusDb({name:ext.Name}).get();
				status = statusRec[0].status;
				klass = "cannotfind";
				if(status == 1) klass = "isDefault";
				if(status == 2) klass = "isConfigured";
			}else{
				klass = "cannotfind";
			}
			item.className = klass;
			theList.appendChild(item);
		}
		
		// Add the user-defined names to the end
		let names = this.UserDefinedNamesDb().order('udname').get();
		for(let i in names){
			let name = names[i].udname;
			let item = document.createElement('menuitem');
			item.setAttribute('label',name);
			item.setAttribute('value',name);
			item.className = 'userdefined';
			item.addEventListener('click', function(event){OPIE2.dispExtInfo(event,this)}, false);
			
			let cnt = this.ExtPrefsDb({ename:name}).count();
			let klass, status;
			if(cnt != 0){
				let cnt2 = this.ExtensionStatusDb({name:name}).count();
				if(cnt2 == 0){
					let obj = {};
					obj.name = name;
					obj.guid = "userdefined";
					obj.status = 2;
					this.ExtensionStatusDb.insert(obj);	
				}
				let statusRec = this.ExtensionStatusDb({name:name}).get();
				status = statusRec[0].status;
				klass = "cannotfind";
				if(status == 1) klass = "isDefault";	// Should never happen?
				if(status == 2) klass = "isConfigured";
			}else{
				klass = "cannotfind";
			}
			item.className += " "+klass;
			theList.appendChild(item);
		}
		return true;
	},

	dispExtInfo: function(event,that){
		let ename = that.label;
		let ext = this.InstalledExtsDb({Name:ename}).get();
		let isUserDefined = (ext[0].Guid == "userdefined") ? true : false;
		let d1 = document.getElementById("extName");
		d1.value = ext[0].Name;
		d1.hidden = false;
		if(!isUserDefined){
			let d2 = document.getElementById("extVersionLabel");
			d2.hidden = false;
			let d3 = document.getElementById("extVersion");
			d3.value = ext[0].Version;
			d3.hidden = false;
			let d4 = document.getElementById("extGuidLabel");
			d4.hidden = false;
			let d5 = document.getElementById("extChromenameLabel");
			d5.hidden = false;
			let d6 = document.getElementById("extChromename");
			d6.value = ext[0].chromename;
			d6.hidden = false;
		}else{
			let d2 = document.getElementById("extVersionLabel");
			d2.hidden = true;
			let d3 = document.getElementById("extVersion");
			d3.hidden = true;
			let d4 = document.getElementById("extGuidLabel");
			d4.hidden = true;
			let d5 = document.getElementById("extChromenameLabel");
			d4.hidden = true;
			let d6 = document.getElementById("extChromename");
			d6.hidden = true;
		}
		let d7 = document.getElementById("extGuid");
		if(!isUserDefined){
			d7.value = ext[0].Guid;	
		}else{
			d7.value = this.opie2Msg[43];;	
		}
		d7.hidden = false;
		
		event.stopPropagation();
		let button = document.getElementById("configExtDropdownButtonID");
		let newButtonLabel = document.getElementById("opie2.config.getprefs").value;
		button.label = newButtonLabel;
		
		let d8 = document.getElementById("resetBoxID");
		d8.hidden = false;
		return true;
	},
		
	buildExtPrefTree: function(){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		let ext = document.getElementById("extName").value;
		let theList = document.getElementById("extPrefList");
		
		// Remove the existing list and rebuild it
		let children = theList.childNodes;
		let n = children.length;
		for (let i = n-1; i >= 0; i--){
			let node = children[i];
			if(node.tagName == 'listitem') theList.removeChild(node);
		}
		this.opie2ExtList = [];
		let cnt = 0;
		let prefs = this.ExtPrefsDb({ename:ext}).order('pname').get();
		for(let i in prefs){
			let pref = prefs[i];
			let row = document.createElement('listitem');
			let cell = document.createElement('listcell');
			cell.setAttribute('label',pref.pname);
			cell.setAttribute('id',pref.pname);
			row.appendChild(cell);
		
			cell = document.createElement('listcell');
			cell.setAttribute('label',pref.ptype);
			row.appendChild(cell);
			row.id = pref.pname;
			theList.appendChild(row);
			this.opie2ExtList.push(pref.pname);
			cnt++;
		}
		document.getElementById("assignedPrefsCnt").value = cnt;
		let filter = document.getElementById("fxPrefFilter");
		this.fxPrefFilter(filter);
		return true;
	},

	deleteAssignedPref: function(event,that,checkForKeypress){
		// ExtPrefsDb: ename, guid, pname, ptype
		// ExtensionStatusDb: guid, name, status
		if(that.selectedIndex == -1) return true
		let nodes = that.getElementsByTagName('listitem');
		let ok = false;
		if(checkForKeypress){
			if(event.keyCode == event.DOM_VK_DELETE) ok=true;
		}else{
			ok = true;
		}
		if(ok){
			let selectedItems = that.selectedItems;
			for(let i=selectedItems.length-1; i>=0; i--){
				let node = selectedItems[i];
				let pref = node.childNodes[0].id;
				that.removeChild(node);
				this.ExtPrefsDb({pname:pref}).remove();
				for(let j in this.opie2ExtList){
					if(this.opie2ExtList[j] == pref) this.opie2ExtList.splice(j,1);
				}
				let cnt = document.getElementById("assignedPrefsCnt").value - 1;
				document.getElementById("assignedPrefsCnt").value = cnt;
			}
			let filter = document.getElementById("fxPrefFilter");
			this.fxPrefFilter(filter);
		}
		return true;
	},

	insertAssignedPref: function(event,that,checkForKeypress){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status
		
		if(that.selectedIndex == -1) return true
		let ok = false;
		if(checkForKeypress){
			if(event.keyCode == event.DOM_VK_INSERT) ok=true;
		}else{
			ok = true;
		}
		if(ok){
			let selectedItems = that.selectedItems;
			for(let i=selectedItems.length-1; i>=0; i--){
				let node = selectedItems[i];
				let pref = node.id;
				pref = pref.substr(3);
				that.removeChild(node);
				let ext = this.FxPrefsDb({name:pref}).get();
				for(let j in this.opie2FxPrefList){
					if(this.opie2FxPrefList[j] == pref) {
						this.opie2FxPrefList.splice(j,1);
					}
				}
				let name = document.getElementById("extName").value;
				let rec = this.InstalledExtsDb({Name:name}).get();
				eInfo = rec[0];
				let item = {};
				item.ename = name;
				item.guid = eInfo.Guid;
				item.pname = ext[0].name;
				item.ptype = ext[0].type;
				this.ExtPrefsDb.insert(item);
				let cnt = document.getElementById("availablePrefsCnt").value - 1;
				document.getElementById("availablePrefsCnt").value = cnt;
			}
			this.buildExtPrefTree();
		}
		return true;
	},

	resetPref: function(){
		// Reset the prefs for an individual extension
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status
		
		let ext = document.getElementById("extName").value;
		let guid = document.getElementById("extGuid").value;
		
		// Clear existing prefs
		this.ExtPrefsDb({ename:ext}).remove();
		
		// Write default prefs
		let exts = this.InstalledExtsDb({Name:ext}).get();
		let rec = exts[0];
		let typeAll = [];
		let type1 = [];
		let type2 = [];
		let type3 = [];
		let type4 = [];
		
		let type1checked = this.opie2Prefs.getBoolPref("extensions.opie2.type1");	
		let type2checked = this.opie2Prefs.getBoolPref("extensions.opie2.type2");		
		let type3checked = this.opie2Prefs.getBoolPref("extensions.opie2.type3");
		let type4checked = this.opie2Prefs.getBoolPref("extensions.opie2.type4");	
		let allChecked = (type1checked && type2checked && type3checked && type4checked);
		
		if(type1checked) type1 = this.GetType1(rec);
		if(type2checked) type2 = this.GetType2(rec);
		if(type3checked) type3 = this.GetType3(rec);
		if(type4checked) type4 = this.GetType4(rec);
		
		if(allChecked){
			this.ExtensionStatusDb({guid:guid}).update({status:1});	// Set status to 'default' (status = 1)
		}else{
			this.ExtensionStatusDb({guid:guid}).update({status:2});	// Set status to 'changed' (status = 2)
		}

		typeAll = typeAll.concat(type1,type2,type3,type4);
		for(let i in typeAll){
			let item = typeAll[i];			
			if(this.ExtPrefsDb(item).count() == 0) this.ExtPrefsDb.insert(item);
		}
		this.buildExtPrefTree();
		let tmp = this.opie2Msg[21].replace("%EXTNAME%",ext);
		this.opie2Alert(tmp,false);
		return true;
	},
	
	opie2MoveButton: function(){
		// If an item in "extPrefList" is selected then move it to "fxPrefList"
		// If an item in "fxPrefList" is selected then move it to "extPrefList"
		// If nothing is selected then return
		let type = 0;
		let exList = document.getElementById("extPrefList");
		let fxList = document.getElementById("fxPrefList");
		if(exList.selectedIndex != -1) type = 1;
		if(fxList.selectedIndex != -1) type = 2;
		let name = document.getElementById("extName").value;

		switch (type){
			case 1:	
				this.deleteAssignedPref(null,exList,false);
				this.enableMoveButton(false);
				break;
			case 2:
				this.insertAssignedPref(null,fxList,false);
				this.enableMoveButton(false);
				this.ExtensionStatusDb({name:name}).update({status:2});
				break;
			default:
				this.enableMoveButton(true);
				return true;
				break;
		}
		
		if(type ==1 || type == 2){
			let cnt = this.ExtensionStatusDb({name:name}).count();
			if(cnt != 0){
				this.ExtensionStatusDb({name:name}).update({status:2});	// Set status to 'cofigured' (status = 2)
			}else{
				let obj = {};
				obj.name = name;
				obj.guid = "userdefined";
				obj.status = 2;
				this.ExtensionStatusDb.insert(obj);	
			}
		}
		return true;
	},
	
	enableMoveButton: function(bool){
		let button = document.getElementById("moveButton");
		button.disabled = !bool;
		return true;
	},
	
	test: function(){
		return true;
	},
	
	buildFxPrefTree: function(filter){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		let theList = document.getElementById("fxPrefList");
		
		// Remove the existing list and rebuild it
		let children = theList.childNodes;
		let n = children.length;
		for (let i = n-1; i >= 0; i--){
			let node = children[i];
				if(node.tagName == 'listitem') theList.removeChild(node);
		}
		let prefs;
		this.opie2FxPrefList = [];
		if(typeof filter == 'undefined'){
			prefs = this.FxPrefsDb().order('name').get();
		}else{
			prefs = this.FxPrefsDb({name:{"likenocase":filter}}).order('name').get();
			//prefs = this.FxPrefsDb({name:{"regex":filter}}).order('name').get();
		}
		let cnt = 0;
		for(let i in prefs){
			let pref = prefs[i];
			let ok = true;
			for(let j in this.opie2ExtList){
				// Don't display items that are already assigned to the extension
				if(this.opie2ExtList[j] == pref.name){
					ok = false;
					break;
				}
			}
			if(!ok) continue;
			let row = document.createElement('listitem');
			let cell = document.createElement('listcell');
			cell.setAttribute('label',pref.name);
			row.appendChild(cell);
			
			cell = document.createElement('listcell');
			cell.setAttribute('label',pref.type);
			row.appendChild(cell);
			row.setAttribute('class','unassigned');
			
			row.id = "Fx:"+pref.name;
			theList.appendChild(row);
			cnt++;
			this.opie2FxPrefList.push(pref.name);
		}
		document.getElementById("availablePrefsCnt").value = cnt;
		return true;
	},
		
	fxPrefFilter: function(that){
		let filter = that.value;
		this.buildFxPrefTree(filter);
		return true;
	},
	
	buildUDnameTree: function(){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		let theList = document.getElementById("UserdefinedList");
		
		// Remove the existing list and rebuild it
		let children = theList.childNodes;
		let n = children.length;
		for (let i = n-1; i >= 0; i--){
			let node = children[i];
				if(node.tagName == 'listitem') theList.removeChild(node);
		}
		
		this.opie2UDList  = [];
		let names = this.UserDefinedNamesDb().order('udname').get();
		let cnt = 0;
		for(let i in names){
			let name = names[i];
			let row = document.createElement('listitem');
			let cell = document.createElement('listcell');
			cell.setAttribute('label',name.udname);
			row.appendChild(cell);
			row.id = name.udname;
			theList.appendChild(row);
			this.opie2UDList.push(name.udname);
			cnt++;
		}
		document.getElementById("userdefinedcount").value = cnt;
		return true;
	},
	
	addUDname: function(){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status
		
		let name = document.getElementById("userdefinedNameID").value;
		let ok = true;
		if(name.length == 0) ok = false;
		if(this.UserDefinedNamesDb({udname:name}).count() != 0) ok = false;
		if(!ok){
			this.opie2PlaySound('failure');
			return false;
		}
		
		let item  = {};
		item.udname = name;
		this.UserDefinedNamesDb.insert(item);
		
		item = {};
		item.Guid = "userdefined";
		item.Name = name;
		item.DatafileName = "";
		item.Version = "";
		item.isXPI = false;
		item.chromename = "";
		this.InstalledExtsDb.insert(item);	
		let button= document.getElementById("includeUDchkbox");
		button.disabled = false;

		this.initConfigMenu();
		this.buildUDnameTree();
		return true;
	},
	
	deleteUDname: function(event,that){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		if(that.selectedIndex == -1) return true
		let nodes = that.getElementsByTagName('listitem');
		if(event.keyCode == event.DOM_VK_DELETE){
			let selectedItems = that.selectedItems;
			for(let i=selectedItems.length-1; i>=0; i--){
				let node = selectedItems[i];
				let name = node.id;
				that.removeChild(node);
				this.UserDefinedNamesDb({udname:name}).remove();
				for(let j in this.opie2UDList){
					if(this.opie2UDList[j] == name) this.opie2UDList.splice(j,1);
				}
				this.InstalledExtsDb({Name:name}).remove();
				this.ExtPrefsDb({ename:name}).remove();
				this.ExtensionStatusDb({name:name}).remove();
				let cnt = document.getElementById("userdefinedcount").value - 1;
				document.getElementById("userdefinedcount").value = cnt;
				let button= document.getElementById("includeUDchkbox");
				button.disabled = (cnt == 0) ? true : false;
			}
		}
		this.initConfigMenu();
		return true;
	},
			
	opie2Close: function(){
		// Save data and close
		this.writeData();
		return true;
	},
	
	setIsXpi: function(item){
		// If the extension location is a directory, it is not an .xpi file
		let id = item.Guid;
		let curExtDir = this.opie2ExtDir.clone();
		curExtDir.append(id);
		let isXpi;
		if(curExtDir.exists() && curExtDir.isDirectory()){
			isXpi = false;
		}else{
			isXpi = true;
		}
		return isXpi;		
	},

	exportPrefs: function (){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status
		
		let tmp, style;
		let cnt = this.ExtPrefsDb().count();
		if(cnt == 0){
			tmp = this.opie2Msg[44];
			this.opie2Alert(tmp);
			return true;
		}
		
		// Get and verify export directory
		if(!this.getExportDir()){return false;}
		
		// Get 'combined' file name
		let type = this.opie2Prefs.getCharPref('extensions.opie2.exportOption'); 
		let combined = (this.opie2Prefs.getCharPref('extensions.opie2.exportOption') == "combined") ? true : false; ;
		let name = document.getElementById("opie2ExportCombinedFilenameID").value; 
		if(combined){
			if(name.length == 0){
				let style = "<style>color: red;font-size: 10pt;</style>";
				let tmp = style+this.opie2Msg[16]; 
				this.opie2Alert(tmp);	// No name for combined export
				return false;
			}		
		}
		
		// Get overwrite option
		this.opie2Overwrite = this.opie2Prefs.getBoolPref('extensions.opie2.overwrite');
		
		// Build an array of the individual preferences
		let numSelected = 0; 
		let combinedList = []; 
		let extCnt = 0; 
		for(let i in this.opie2ExtNames){
			let id = "expList:"+this.opie2ExtNames[i];
			let listItem = document.getElementById(id); 
			let isChecked = (listItem.getAttribute("checked") == "true") ? true : false; 
			if(!isChecked) continue;
			let item = this.opie2ExtNames[i];
			let prefs = this.ExtPrefsDb({ename:item}).get();
			let prefList = [];
			for(let i in prefs){
				prefList.push(prefs[i].pname);
			}
			if(prefs.length == 0){
				tmp = this.opie2Msg[12].replace("%EXT%",item);
				this.opie2Alert(tmp,true);
				continue;
			}
			numSelected++;
			if(combined){
				combinedList = combinedList.concat(prefList);
				extCnt++;
			}else{
				if(this.opie2Export(this.opie2ExportDir.path,item,prefList)){extCnt++;}
			}
		}

		if(numSelected == 0){
			style = "<style>color: red;font-size: 10pt;</style>";
			tmp = style+this.opie2Msg[02]; 
			this.opie2Alert(tmp);	// No extensions selected
			return false;
		}
		
		style = "<style>color: blue;font-size: 10pt;</style>";
		if(combined){
			name += ".oprefs";
			let ok = this.opie2ExportCombined(this.opie2ExportDir.path,name,combinedList); 
			if(ok){
				tmp = this.opie2Msg[17].replace(/%NUM%/,extCnt); 
				tmp = style+tmp.replace(/%NAME%/,name);
			}else{
				tmp = this.opie2Msg[15].replace(/%NAME%/,name); 
				let style = "<style>color: red;font-size: 10pt;</style>";
				tmp = style+tmp;
			}
		}else{
			tmp = style+this.opie2Msg[06].replace(/%NUM%/,extCnt); 
		}
		if(this.opie2Err != 4){this.opie2Alert(tmp);}
		this.clearExportList();
		this.buildExportList();
		this.opie2ErrCnt = 0;
		return true;
	},
	
	readyTmpDir: function(){
		// Ready a tmp directory.  If it already exist, clear it out.  Otherwise created it.
		this.opie2TmpDir = this.opie2FileUtils.getDir("TmpD", [], true);
		this.opie2TmpDir.append("opie2Tmp");

		if(this.opie2TmpDir.exists()){
			entries = this.opie2TmpDir.directoryEntries;
			while(entries.hasMoreElements()){
				entry = entries.getNext();
				entry.QueryInterface(Ci.nsIFile);
				let file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
				file.initWithPath(entry.path);
				try{
					if(file.exists()) file.remove(false);
				}catch(err){
					let style =  "<style>color: red;font-size: 10pt;</style>";
					let tmp = style+this.opie2Msg[10]; 
					tmp += "\n"+this.opie2TmpDir.path+"\n";
					tmp += this.opie2Msg[11]+"\n\n"; 
					tmp += err.message;
					this.opie2Alert(tmp,true);
					return false;
				}
			}
		}else{
			try{
				this.opie2TmpDir.create(Ci.nsIFile.DIRECTORY_TYPE, this.opie2FileUtils.PERMS_DIRECTORY);
			}catch(err){
				let style =  "<style>color: red;font-size: 10pt;</style>";
				let tmp = style+this.opie2Msg[33]; 
				tmp += "\n"+this.opie2TmpDir.path+"\n";
				tmp += this.opie2Msg[34]+"\n\n"; 
				tmp += err.message;
				this.opie2Alert(tmp,true);
				return false;
			}
		}
		return true;
	},

	clearExportList: function (){
		let theTree = document.getElementById("opie2ExportList"); 
		let children = theTree.childNodes; 
		let n = children.length; 
		
		// Clear existing items
		for (let i=0; i < n;  i++) {
			theTree.removeChild(children[0]);
		}
		return true;
	},

	buildExportList: function (){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		let theList = document.getElementById("opie2ExportList");
		this.opie2SelectionCount = 0;
		this.opie2ExtensionCount = 0;
		this.opie2ExtNames = [];
		
		let exts = this.InstalledExtsDb().order("Name").get();	// An array of objects sorted by name
		let udList = [];
		for(let i in exts){
			let ext = exts[i];
			let prefCnt = this.ExtPrefsDb({guid:ext.Guid}).count();			
			if(prefCnt == 0) continue;
			if(ext.Guid == "userdefined"){
				udList.push(ext);
				continue;
			}
			let row = document.createElement('listitem'); 
			row.id = "expList:"+ext.Name;
			row.setAttribute('type', "checkbox");
			row.setAttribute('label', ext.Name+" (v"+ext.Version+")");
			row.setAttribute('checked', "false");
			row.onclick = function(event){OPIE2.selectCount(this.checked)};
			theList.appendChild(row);
			this.opie2ExtNames.push(ext.Name);
			this.opie2ExtensionCount++;
		}
		
		// Add user-defined exts at the bottom
		for(let i in udList){
			let ext = udList[i];
			let row = document.createElement('listitem'); 
			row.id = "expList:"+ext.Name;
			row.setAttribute('type', "checkbox");
			row.setAttribute('label', ext.Name+" ("+this.opie2Msg[43]+"){"+this.opie2ProfName+"}");
			row.setAttribute('checked', "false");
			row.className = "fontitalic";;
			row.onclick = function(event){OPIE2.selectCount(this.checked)};
			theList.appendChild(row);
			this.opie2ExtNames.push(ext.Name);
			this.opie2ExtensionCount++;
		}
		this.setExportCaption();
		if(this.opie2ExtensionCount == 0) this.opie2Alert(this.opie2Msg[03]);
		return true;
	},
		
	selectCount: function(checked){
		if(checked){
			this.opie2SelectionCount++;
		}else{
			this.opie2SelectionCount--;
		}
		this.setExportCaption();
	},
	
	setExportCaption: function(){
		let d = document.getElementById('extcount');
		let value = this.opie2Msg[20].replace(/%CNT%/,this.opie2ExtensionCount);
		value = value.replace(/%SEL%/,this.opie2SelectionCount);
		d.value = value;
		return true;
	},

	opie2Export: function(dir,name,prefList){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		// Build export data from 'prefList' and write it to 'which.DatafileName' located in 'dir'
		let ext = this.InstalledExtsDb({Name:name}).get();
		let which = ext[0];
		let opie2DataFile = Cc["@mozilla.org/file/local;1"]
				.createInstance(Ci.nsIFile);
		opie2DataFile.initWithPath(dir);
		opie2DataFile.append(which.DatafileName);

		if(!this.opie2Overwrite && opie2DataFile.exists()){return false;}

		let dataFile = Cc["@mozilla.org/network/file-output-stream;1"]
			.createInstance(Ci.nsIFileOutputStream);
		dataFile.init(opie2DataFile, 0x02 | 0x08 | 0x20, 0755, 0); // write, create, truncate
		
		let charset = "UTF-8"; 
		let os = Cc["@mozilla.org/intl/converter-output-stream;1"]
			 .createInstance(Ci.nsIConverterOutputStream);

		os.init(dataFile, charset, 4096, 0x0000);
		let list = this.opie2BuildExportFileData(prefList); 
		if(list != ""){os.writeString(list);}	
		os.close();
		if(this.opie2Err == 4){return false;}
		return true;
	},

	opie2ExportCombined: function(dir,fname,prefList){
		let opie2DataFile = Cc["@mozilla.org/file/local;1"]
			.createInstance(Ci.nsIFile);
		opie2DataFile.initWithPath(dir);
		opie2DataFile.append(fname);
		if(!this.opie2Overwrite && opie2DataFile.exists()){return false;}

		let dataFile = Cc["@mozilla.org/network/file-output-stream;1"]
			.createInstance(Ci.nsIFileOutputStream);
		dataFile.init(opie2DataFile, 0x02 | 0x08 | 0x20, 0755, 0); // write, create, truncate
		
		let charset = "UTF-8"; 
		let os = Cc["@mozilla.org/intl/converter-output-stream;1"]
			 .createInstance(Ci.nsIConverterOutputStream);

		os.init(dataFile, charset, 4096, 0x0000);
		let list = this.opie2BuildExportFileData(prefList); 
		if(list != ""){os.writeString(list);}	

		os.close();
		return true;
	},

	getExportDir: function (){
		// Get and verify export directory
		let dir = document.getElementById("opie2ExportDirID").value;
		if(dir == ""){
			let style = "<style>color: red;font-size: 10pt;</style>";
			let tmp = style+this.opie2Msg[1]; 
			this.opie2Alert(tmp); // No install directory	
			return false;
		}
		
		// Verify export directory
		let file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);                 
		let err = false; 
		try {
			file.initWithPath(dir);
		}catch(e){err = true;}
		if (!(file.exists() && file.isDirectory()) || err){
			let style = "<style>color: red;font-size: 10pt;</style>";
			let tmp = style+this.opie2Msg[05].replace(/%DIR%/,dir); 
			this.opie2Alert(tmp);
			return false;
		}
		this.opie2ExportDir = file;
		return true;
	},

	opie2BuildExportFileData: function(prefList){
		// Return export data built from 'prefList'
		let prefService = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService); 
		let prefBranch = prefService.getBranch(null); 		
		let fileData = "";
		for(let i=0; i<prefList.length; i++){
			let pref = prefList[i];
			let type = prefBranch.getPrefType(pref);
			let value = this.opie2GetExtPref(pref,type);
			
			switch(this.opie2Err){
				case 0:	// No error
					break;
				case 1:	// Skip this preference and continue exporting
					continue;
					break;
				case 2:	// Skip this extension and continue exporting
					return "";
					break;
				case 3:	// Always ignore this error (can be changed in OPIE2 options)
					this.opie2Prefs.setBoolPref("extensions.opie2.ignoreError", true);
					continue;
					break;
				case 4:	// Abort this export session
					var style = "<style>color: red;font-size: 10pt; font-weight: bold;</style>";
					let tmp = style+this.opie2Msg[28]+"\n"; 
					style = "<style>color: black;font-size: 10pt; font-weight: normal;</style>";
					tmp += style+this.opie2Msg[29];
					this.opie2Alert(tmp);
					return "";
					break;
			}

			let Type;
			let prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefBranch);
			switch (type){
				case prefs.PREF_BOOL:	
					Type = "boolean";
					break;
				case prefs.PREF_INT:
					Type = "integer";
					break;
				case prefs.PREF_STRING:
					Type = "string";
					break;
			}
			let line = new this.opie2PrefObj(Type,pref,value); 
			fileData += JSON.stringify(line) + "\n";
		}
		return fileData;
	},
	
	opie2GetExtPref: function(prefName,type){
		// Get the extension preference
		this.opie2Err = 0;
		let opie2ErrCnt = 0;
		try{
			let prefValue;
			switch(type){
				case 32:	// String
					prefValue = this.opie2GetUnicharPref(prefName);
					break;
				case 64:	// Integer
					prefValue = this.opie2Prefs.getIntPref(prefName);
					break;
				case 128:	// Boolean
					prefValue = this.opie2Prefs.getBoolPref(prefName);
					break;
			}
			return prefValue;
		}catch(e){
			opie2ErrCnt++;
			let ignoreError = this.opie2Prefs.getBoolPref("extensions.opie2.ignoreError"); 
			let err = e.toString().indexOf("NS_ERROR_UNEXPECTED"); 
			if(ignoreError && err != -1){
				this.opie2Err = 1;
				return "";
			}else{
				if(err == -1){
					alert(e);	// Non-locking error
					this.opie2Err = 4;
					return "";
				}else{
					this.opie2Alert(e.toString());
				}
			}
		}
	},

	opie2ViewExDir: function (){
		let folder = document.getElementById("opie2ExportDirID").value;
		let nsLocalFile = Components.Constructor("@mozilla.org/file/local;1","nsILocalFile", "initWithPath");
		try{
			new nsLocalFile(folder).reveal();
		}catch(err){
			let tmp = this.opie2Msg[45].replace("%DIR%",folder);
			this.opie2Alert(tmp,false);
		}
		return true;
	},
	
	buttonCheck: function(){
		// ExtPrefsDb: ename, guid, pname, ptype
		// FxPrefsDb:  name, type, value
		// InstalledExtsDb: Guid, Name, DatafileName, Version, isXPI, chromename
		// UserDefinedNamesDb: udname
		// ExtensionStatusDb: guid, name, status

		let folder = document.getElementById("opie2ExportDirID").value;
		let button1 = document.getElementById("opie2viewbudirbtnID");
		let button2 = document.getElementById("opie2clearbudirbtnID");
		let disabled = false;
		if(!folder) disabled = true;
		button1.disabled = disabled;
		button2.disabled = disabled;
		
		// Enable/disable 'Include user-defined packages'
		let button= document.getElementById("includeUDchkbox");
		let cnt = this.InstalledExtsDb({Guid:'userdefined'}).count();
		button.disabled = (cnt == 0) ? true : false;
		button.checked = false;
		return true;
	},

	opie2ClearExDir: function (){
		if(this.opie2Prefs.getBoolPref('extensions.opie2.warnbeforeclear')){
			this.opie2ConfirmClear();
			if(!this.opie2Prefs.getBoolPref('extensions.opie2.confirm')) return false;
		}
		
		if(!this.getExportDir()){return false;}

		let entries = this.opie2ExportDir.directoryEntries; 
		while(entries.hasMoreElements()){
			let entry = entries.getNext(); 
			entry.QueryInterface(Ci.nsIFile);
			if(entry.isFile()){entry.remove(false);}
		}
		let style = "<style>color: blue;font-size: 10pt;</style>";
		let tmp = style+this.opie2Msg[14]; 
		this.opie2Alert(tmp);
		return true;
	},

	opie2ExpandAll: function(){
		var tree = document.getElementById("opie2InitAllTree");
		var treeView = tree.treeBoxObject.view;
		for (var i = 0; i < treeView.rowCount; i++) {
			if (treeView.isContainer(i) && !treeView.isContainerOpen(i)){
				treeView.toggleOpenState(i);
			}
		}
		return true;
	},

	opie2CollapseAll: function (){
		var tree = document.getElementById("opie2InitAllTree");
		var treeView = tree.treeBoxObject.view;
		for (var i = 0; i < treeView.rowCount; i++) {
			if (treeView.isContainer(i) && treeView.isContainerOpen(i)){
				treeView.toggleOpenState(i);
			}
		}
		return true;
	},

	opie2CountPreferences: function(prefRoot){
		// Count the number of about:config entries (preferences)
		let prefCount = {value:0}; 
		let prefService = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService); 

		if(prefRoot === undefined) return 0;
		let prefBranch = prefService.getBranch(prefRoot); 
		let prefArray = prefBranch.getChildList(".", prefCount);
		return prefCount.value;
	},
		
	opie2Help: function(helpIndex,fieldName){
		let opts = this.opie2SetWinOpts(false);
		this.opie2Win = window.openDialog('chrome://opie2/content/opie2Help.xul', 'OPIE2 Help', opts,helpIndex,fieldName);
		this.opie2Win.focus();
		return true;
	},

	opie2Alert: function(message,modal){
		let opts = this.opie2SetWinOpts(modal);
		this.opie2Win = window.openDialog('chrome://opie2/content/opie2Alert.xul','OPIE2 Alert',opts,message);	
		this.opie2Win.focus();
		return true;
	},

	opie2ConfirmClear: function (){
		let opts = this.opie2SetWinOpts(true);
		this.opie2Win = window.openDialog('chrome://opie2/content/opie2ConfirmClear.xul','OPIE2 Confirm',opts);	
		this.opie2Win.focus();
		return true;
	},
	
	setConfirmValue: function(bool){
		this.opie2Prefs.setBoolPref('extensions.opie2.confirm',bool);
	},
	
	setConfirmMsg: function (){
		this.opie2SetMsgs();
		let msgBox = document.getElementById("opie2MsgBox");
		let labelElm = document.createElement('label');
		let label = labelElm.cloneNode(false);
		label.value = this.opie2Msg[29];
		label.className = 'fontwarn';
		msgBox.appendChild(label);
		label = labelElm.cloneNode(false);
		label.value = this.opie2Msg[30];
		label.className = 'fontnormal';
		msgBox.appendChild(label);
		OPIE2.opie2PlaySound('beep');
		return true;
	},

	opie2SetWinOpts: function(modal){
		let opts = "chrome,dialog,alwaysRaised,resizable";
		if(modal){opts += ",modal";}
		let opie2MsgType = this.opie2Prefs.getCharPref('extensions.opie2.alertType');
		if(opie2MsgType == 'slider'){opts += ",popup";}
		if(opie2MsgType == 'standard'){opts += ",centerscreen,titlebar";}
		return opts;
	},

	opie2SetAlertMsg: function (){
		// If the message contains a style or class, it must be the first part of the message 
		// (one style per line, if needed).  
		// If a class is assigned, it must be in the form of: <class>classname</class>
		let tmp = window.arguments[0]+"   \n\n";
		let lines = tmp.split(/\n/); // Break it into an array
		let msg = [], re, a;
		let separator = document.createElement('separator');
		for(let i=0;i<lines.length;i++){	
			let line = lines[i];
			let item = {};
			item.Style = "font-family: arial; font-size: 10pt; font-weight: normal; color: purple";
			item.Text = line;

			// Parse for links
			let link = line.indexOf("<link>");
			if(link == 0){
				re = /(<link>)(.*),(.*)(<\/link>)/g
				a = re.exec(line);
				item.Text = a[2];
				item.Link = a[3];
				item.Klass = "link";	
				msg[i] = item;
				continue;
			}
			
			// Parse for styles
			let style = line.indexOf("<style>");
			if(style == 0){
				re = /(<style>)(.*)(<\/style>)(.*)/g
				a = re.exec(line);
				item.Style = a[2];
				item.Text = a[4];
				msg[i] = item;
				continue;
			}
			
			// Parse for class
			let klass = line.indexOf("<class>");
			if(klass == 0){
				re = /(<class>)(.*)(<\/class>)(.*)/g
				a = re.exec(line);
				item.Klass = a[2];
				item.Text = a[4];
				msg[i] = item;
				continue;
			}
			
			msg[i] = item;
		}
		this.opie2Clipboard = [];
		let msgBox = document.getElementById("opie2MsgBox");
		if(msgBox){
			for(let i=0; i<msg.length; i++){
				if(msg[i].Text == ""){
					msgBox.appendChild(separator.cloneNode(false));
					this.opie2Clipboard.push(msg[i].Text);
					continue;
				}
				let desc = document.createElement('description');
				desc.setAttribute("style",msg[i].Style);
				if(msg[i].Klass){
					desc.removeAttribute("style");
					desc.setAttribute("class",msg[i].Klass);
					if(msg[i].Klass == "link"){
						let txt = document.createElement('text');
						txt.setAttribute("value",msg[i].Text);
						txt.removeAttribute("onclick");
						clickify(txt, msg[i].Link);
						desc.appendChild(txt);
					}else{
						desc.textContent = msg[i].Text;			
					}
				}else{
					desc.textContent = msg[i].Text;			
				}
				msgBox.appendChild(desc);
				this.opie2Clipboard.push(msg[i].Text);
			}
			try{
				document.documentElement.getButton("cancel").focus();
			}catch(err){;;}
		}
		this.opie2PlaySound('message');
		
		function clickify(elem, address) {
			elem.setAttribute("href", address);
			elem.onclick = function(event){OPIE2.opie2OpenLink(this.getAttribute('href'))};
			return true;
		}
		return true;
	},
	
	opie2OpenLink: function(URL){
		if(!URL){return false;}
		let wm = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
		let win = wm.getMostRecentWindow("navigator:browser");
		let currBlank = false;
		try{
			let opie2GetBrowser = win.getBrowser();
			let loc = opie2GetBrowser.mCurrentTab.linkedBrowser.contentDocument.location;
			if(loc == "about:blank") currBlank = true;
			if(loc == "about:newtab") currBlank = true;
			if(opie2GetBrowser.mCurrentTab.label == "(Untitled)") currBlank = true;
		}catch(e){;;}
		if(currBlank){
			win.loadURI(URL);
		}else{
			win.openNewTabWith(URL, this.docURL, null, null);
		}
		return win;
	},

	opie2SetAlertMsg: function (){
		this.opie2SetConfirmMsg();
		return true;
	},
		
	opie2SetConfirmMsg: function (){
		// If the message contains a style, it must be the first part of the message (one style per line, if needed)
		let tmp = window.arguments[0]+"   \n\n"; 
		let msg = {}; 
		let i = 0; 
		while(true){	// Break it into an array
			let item = new this.opie2AlertObj(); 
			let p = tmp.indexOf("\n"); 
			if(p == -1){break;}
			item.Style = "font-family: arial; font-size: 10pt; font-weight: normal; color: purple";  // Default style
			let ps1 = tmp.indexOf("<style>"); 
			if(ps1 == 0){	
				let ps2 = tmp.indexOf("</style>"); 
				if(ps2 != -1){
					item.Style = tmp.substr(7,ps2-7);			// "7" equals the length of "<style>"
					tmp = tmp.replace(tmp.substr(0,ps2+8),"");	// Remove the style
					p = tmp.indexOf("\n");
				}
			}
			item.Text = tmp.substr(0,p);
			msg[i] = item;
			i++;
			tmp = tmp.substr(p+1);	
		}
		this.opie2Clipboard = [];

		let msgBox = document.getElementById("opie2MsgBox"); 
		if(msgBox){
			for(let i in msg){
				let desc = document.createElement("description"); 
				desc.textContent = msg[i].Text;
				desc.setAttribute("style",msg[i].Style);
				msgBox.appendChild(desc);
				this.opie2Clipboard.push(msg[i].Text);
			}
		}
		try{		
			let dialog = document.documentElement; 
			let button = dialog.getButton('extra1');
			button.hidden = false;
		}catch(err){;;}
		this.opie2PlaySound('alert');
		return true;
	},
	
	opie2SliderLoad: function(noSlide){
		let opie2MsgType = OPIE2.opie2Prefs.getCharPref('extensions.opie2.alertType');
		
		// 'noSlide' is for resizable windows opened with window.openDialog()
		if(typeof noSlide == 'undefined') noSlide = false;
		if(noSlide) febeMsgType = "standard";
		sizeToContent();
		let finalHeight = window.outerHeight;
		let windowtype = document.documentElement.getAttribute("windowtype");

		if(opie2MsgType == 'slider'){
			var sliderIncrement = 10;		// Amount the slider grows by (pixels)
			var sliderTime = 25;				// Time between increments (milliseconds)
			window.outerHeight = 1;			// First shrink the window down
			let x =(screen.availLeft + screen.availWidth - window.outerWidth) - 10;
			if(windowtype == 'opie2:help') x=0;	// Help screens popup on the left
			let y = screen.availTop + screen.availHeight - window.outerHeight;
			window.moveTo(x,y);
			window.focus();
			setTimeout(function(){animateAlert();}, sliderTime);
		}
		
		function animateAlert(){
			if(window.outerHeight < finalHeight){
				window.screenY -= sliderIncrement;
				window.outerHeight += sliderIncrement;
				setTimeout(function(){animateAlert()}, sliderTime);
				return true;
			}
		}
	
		// On windows that don't have a 'Copy to clipboard' button, the button label will be blank
		let copy2clipboardBtn = document.documentElement.getButton("extra1");
		let copy2clipboard = OPIE2.opie2Prefs.getBoolPref("extensions.opie2.copyToClipboard");
		copy2clipboardBtn.hidden = (!copy2clipboard || (copy2clipboardBtn.label == ""));
		try{
			document.documentElement.getButton("cancel").focus();
		}catch(err){;;}
		return true;
	},

	opie2CopyToClipboard: function (){
		// Copy the contents of a opie2Alert, opie2Confirm, or opie2Err box to the system clipboard
		let gClipboardHelper = Cc["@mozilla.org/widget/clipboardhelper;1"].getService(Ci.nsIClipboardHelper); 
		gClipboardHelper.copyString(this.opie2Clipboard.join("\n"));
		this.opie2PlaySound('success');
		return true;
	},

	opie2LoadHelp: function (){
		this.opie2SetMsgs(); 
		let index = window.arguments[0]; 
		let field = window.arguments[1]; 
		const HTMLNS = "http://www.w3.org/1999/xhtml";

		document.getElementById("field").value = field;
		let fh = document.getElementById('opie2-help'); 
		let helpTxt = ""; 
		
		// XUL pallette
		let descElm = document.createElement('description'); 
		let separatorElm = document.createElement('separator'); 
		separatorElm.setAttribute('class', "thin");
		let vboxElm = document.createElement('vbox'); 
		let hboxElm = document.createElement('hbox'); 
		let gridElm = document.createElement('grid'); 
		let columnsElm = document.createElement('columns'); 
		let columnElm = document.createElement('column'); 
		let rowsElm = document.createElement('rows'); 
		let rowElm = document.createElement('row'); 
		let spacerElm = document.createElement('spacer'); 
		spacerElm.setAttribute('flex', "1");
		let imageElm = document.createElement('image'); 
		let labelElm = document.createElement('label'); 
		let textElm = document.createElement('text'); 
		
		// HTML namespace elements
		let htmlulElm = document.createElementNS(HTMLNS, "html:ul"); 
		let htmlolElm = document.createElementNS(HTMLNS, "html:ol"); 
		let htmlliElm = document.createElementNS(HTMLNS, "html:li"); 
		let helpTxtbox = document.getElementById('helptextbox'); 
		let desc, text, hbox, listHead, listItem;

		switch(index){
			case "03":
				for(let i=1; i<=2; i++){ 
					helpTxt = fh.getString("opie2.help_"+index+".text"+i);
					this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				}
				helpTxt = fh.getString("opie2.help_"+index+".text3a");
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,false,"");
				
				vbox = vboxElm.cloneNode(false); 
				listHead = htmlulElm.cloneNode(false); 
				
				listItem = htmlliElm.cloneNode(false); 
				helpTxt = fh.getString("opie2.help_"+index+".text3b");
				text = textElm.cloneNode(false); 
				text.setAttribute('value', helpTxt);
				text.setAttribute('class', "isDefault");
				listItem.appendChild(text);
				listHead.appendChild(listItem)

				listItem = htmlliElm.cloneNode(false); 
				helpTxt = fh.getString("opie2.help_"+index+".text3c");
				text = textElm.cloneNode(false); 
				text.setAttribute('value', helpTxt);
				text.setAttribute('class', "isConfigured");
				listItem.appendChild(text);
				listHead.appendChild(listItem)

				listItem = htmlliElm.cloneNode(false); 
				helpTxt = fh.getString("opie2.help_"+index+".text3d");
				text = textElm.cloneNode(false); 
				text.setAttribute('value', helpTxt);
				text.setAttribute('class', "cannotfind");
				listItem.appendChild(text);
				listHead.appendChild(listItem)

				vbox.appendChild(listHead);
				helpTxtbox.appendChild(vbox);
				for(let i=4; i<=7; i++){ 
					helpTxt = fh.getString("opie2.help_"+index+".text"+i);
					this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				}
				break;
			case "04": 
				helpTxt = fh.getString("opie2.help_"+index+".text1");
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"");
										
				text = textElm.cloneNode(false); 
				helpTxt = "1) "+document.getElementById("opie2.export.option.label2").value+":";
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,false,"fontbold");
				
				hbox = hboxElm.cloneNode(false); 
				hbox.setAttribute('flex', "0");
				hbox.setAttribute('class', "helpindent");
				
				text = textElm.cloneNode(false); 
				helpTxt = fh.getString("opie2.help_"+index+".text2");
				text.setAttribute('value', helpTxt);
				text.setAttribute('class', "helpnormal");
				desc = descElm.cloneNode(false); 
				desc.appendChild(text);
				hbox.appendChild(desc);
				
				helpTxt = fh.getString("opie2.help_"+index+".text3");
				desc = descElm.cloneNode(false); 
				desc.setAttribute('class', "fontitalic");
				desc.setAttribute('value', helpTxt);
				hbox.appendChild(desc);
				helpTxtbox.appendChild(hbox);
				
				text = textElm.cloneNode(false); 
				helpTxt = "2) "+document.getElementById("opie2.export.option.label3").value+" ...";
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,false,"fontbold");
				helpTxt = fh.getString("opie2.help_"+index+".text4");
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"helpindent");
				
				helpTxt = document.getElementById("opie2.export.overwrite.chkbox").value;
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,false,"fontbold");
				helpTxt = fh.getString("opie2.help_"+index+".text5");
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"helpindent");

				helpTxt = document.getElementById("opie2.export.ignoreerror.chkbox").value;
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,false,"fontbold");
				helpTxt = fh.getString("opie2.help_"+index+".text6");
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"helpindent");
				
				break;
			case "06": 
				for(let i=1; i<4; i++){ 
					helpTxt = fh.getString("opie2.help_"+index+".text"+i);
					this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				}
				
				hbox = hboxElm.cloneNode(false); 
				hbox.setAttribute('flex', "0");
				
				text = textElm.cloneNode(false); 
				helpTxt = fh.getString("opie2.help_"+index+".text4");
				text.setAttribute('value', helpTxt);
				text.setAttribute('class', "helpnormal");
				desc = descElm.cloneNode(false); 
				desc.appendChild(text);
				hbox.appendChild(desc);
				helpTxtbox.appendChild(hbox);
				
				helpTxt = fh.getString("opie2.help_"+index+".text5");
				desc = descElm.cloneNode(false); 
				desc.setAttribute('class', "fontitalic");
				desc.setAttribute('value', helpTxt);
				hbox.appendChild(desc);
				
				hbox = hboxElm.cloneNode(false); 
				hbox.setAttribute('flex', "0");
				
				text = textElm.cloneNode(false); 
				helpTxt = fh.getString("opie2.help_"+index+".text6");
				text.setAttribute('value', helpTxt);
				text.setAttribute('class', "helpnormal");
				desc = descElm.cloneNode(false); 
				desc.appendChild(text);
				hbox.appendChild(desc);
				
				helpTxt = fh.getString("opie2.help_"+index+".text7");
				desc = descElm.cloneNode(false); 
				desc.setAttribute('class', "fontitalic");
				desc.setAttribute('value', helpTxt);
				hbox.appendChild(desc);
				helpTxtbox.appendChild(hbox);
				break;
			case "07": 
				for(let i=1; i<4; i++){ 
					helpTxt = fh.getString("opie2.help_"+index+".text"+i);
					this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				}
				vbox = vboxElm.cloneNode(false); 
				listHead = htmlulElm.cloneNode(false); 
				for(let i=4; i<14; i++){
					let listItem = htmlliElm.cloneNode(false); 
					helpTxt = fh.getString("opie2.help_"+index+".text"+i);
					text = textElm.cloneNode(false); 
					text.setAttribute('value', helpTxt);
					text.setAttribute('class', "fontitalic");
					listItem.appendChild(text);
					listHead.appendChild(listItem)
				}
				vbox.appendChild(listHead);
				helpTxtbox.appendChild(vbox);
				helpTxt = fh.getString("opie2.help_"+index+".text14");
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				helpTxt = fh.getString("opie2.help_"+index+".text15");
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				break;
			case "08": 
				for(let i=1; i<3; i++){ 
					helpTxt = fh.getString("opie2.help_"+index+".text"+i);
					this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				}
				vbox = vboxElm.cloneNode(false); 
				listHead = htmlolElm.cloneNode(false); 
				for(let i=3; i<7; i++){
					let listItem = htmlliElm.cloneNode(false); 
					helpTxt = fh.getString("opie2.help_"+index+".text"+i);
					text = textElm.cloneNode(false); 
					text.setAttribute('value', helpTxt);
					listItem.appendChild(text);
					listItem.setAttribute('maxwidth', "200");
					listItem.setAttribute('flex', "1");
					listHead.appendChild(listItem)
				}
				vbox.appendChild(listHead);
				helpTxtbox.appendChild(vbox);
				helpTxt = fh.getString("opie2.help_"+index+".text7");
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				helpTxt = fh.getString("opie2.help_"+index+".text8");
				this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				break;
			default:
				let i = 1;
				while(true){ 
					try{
						helpTxt = fh.getString("opie2.help_"+index+".text"+i);
					}catch(err){break}				
					i++;
					this.opie2AddTxtHelpLine(helpTxtbox,helpTxt,true,"");
				}
				break;
		}
		this.opie2PlaySound("help");
		return true;
	},

	opie2AddTxtHelpLine: function(box,txt,separate,klass){
		let desc = document.createElement('description'); 
		desc.setAttribute('flex', "1");
		if(klass != ""){
			desc.setAttribute('class', klass);
		}else{
			desc.setAttribute('class', "helpnormal");
		}
		desc.textContent = txt;
		box.appendChild(desc);
		let separator = document.createElement('separator'); 
		if(separate){
			separator.setAttribute('class', "thin");
			box.appendChild(separator);
		}
	},
	
	pauseThread: function(){
/*		During validation, the following message is generated:
			Banned or deprecated JavaScript Identifier
			Warning: Spinning the event loop with processNextEvent is a common cause of deadlocks, crashes, and other errors due to unintended reentrancy. Please use asynchronous callbacks instead wherever possible
*/				
		// This function is only used when updating xul progress meters.
		// As per Jorge Villalobos, this method is acceptable for this purpose.
		
		let currentThread = Cc['@mozilla.org/thread-manager;1'].getService(Ci.nsIThreadManager).currentThread;
		currentThread.processNextEvent(false);
		return true;
	}
}